/*
 * Copyright (C) 2003 2004 2005, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */


#include <config.h>

#include <glib.h>
#include "sound.h"
#include "player.h"
#include "main.h"
#include "um.h"
#include "rateest.h"

Dataformat player_fallback_format = { DATAFORMAT_PCM, 44100, 2, 0, 0, TRUE, 
				      IS_BIGENDIAN };

/* Info about currently playing sound. */
static ChunkHandle *ch=NULL; /* NULL == stopped */
static off_t loopstart, loopend, curpos;
static off_t realpos_offset;
static gboolean loop, small_loop;
static guint32 zeroes;

/* Output buffer */
static gchar player_buf[4096],small_loop_buf[4096];
static gint player_bufsize,player_bufpos,small_loop_bufsize;



static void get_new_data(void)
{
     off_t frames,sz;
     if (curpos < loopstart) 
	  frames=loopstart-curpos;
     else if (curpos == loopstart && small_loop) {
	  memcpy(player_buf, small_loop_buf, small_loop_bufsize);
	  player_bufsize = small_loop_bufsize;
	  player_bufpos = 0;
	  return;
     } else if (curpos < loopend)
	  frames = loopend-curpos;
     else
	  frames = ch->length - curpos;
     sz = frames * ch->format.samplebytes;
     if (sz > sizeof(player_buf)) sz = sizeof(player_buf);
     /* printf("sz == %d\n",(int)sz); */
     player_bufsize = chunk_read_array(ch,curpos,sz,player_buf);
     player_bufpos = 0;
     curpos += player_bufsize / ch->format.samplebytes;
     if (loop && curpos == loopend) curpos = loopstart;
}

/* Check if the sound driver wants more data and send it.
 * Returns FALSE if it had nothing to do.
 * Since we want to be sure that we don't miss sound at the end, we send some
 * silence before stopping the sound. */
gboolean player_work(void)
{
     guint32 i;
     if (!ch || !output_want_data()) return FALSE;
     if (player_bufpos == player_bufsize) {
	  /* puts("Hey 1"); */
	  get_new_data();
	  /* Still no new data ? Then we should stop.. */
	  if (player_bufpos == player_bufsize) {
	       /* puts("Hey 2"); */
	       /* Emit up to 1/2 second of silence before stopping output. */
	       if (zeroes < ch->format.samplerate*ch->format.samplebytes/2) {
		    memset(player_buf,0,sizeof(player_buf));
		    i = sizeof(player_buf);
		    i -= i%ch->format.samplebytes;
		    zeroes += output_play(player_buf,i);
	       }	       
	       /* Wait until buffers have been emptied and the cursor has 
		* reached the end before stopping */
	       if (output_play(NULL,0)==0 && player_get_real_pos()>=curpos) {
		    output_stop();
		    chunk_close(ch);
		    gtk_object_unref(GTK_OBJECT(ch));
		    ch=NULL;
	       }
	       return FALSE;
	  }
     }
     i = output_play(player_buf+player_bufpos,player_bufsize-player_bufpos);
     player_bufpos += i;
     rateest_log_data(i/ch->format.samplebytes);
     return (i>0);
}


static void check_small_loop(void)
{
     off_t o;
     guint32 u,v;
     small_loop = FALSE;
     if (loop) {
	  o = (loopend-loopstart)*ch->format.samplebytes;
	  if (o <= sizeof(small_loop_buf)) {
	       u = (guint32) o;
	       small_loop = TRUE;
	       chunk_read_array(ch,loopstart,u,small_loop_buf);
	       for (v=1; v < sizeof(player_buf)/u; v++) 
		    memcpy(small_loop_buf+u*v,small_loop_buf,u); 
	       /* printf("check_small_loop: u=%d, v=%d\n",u,v); */
	       small_loop_bufsize = v*u;
	  }
     }
}


gboolean player_play(Chunk *chk, off_t spos, off_t epos, gboolean lp)
{
     gboolean b;
     Dataformat fmt;
     player_stop();
     if (spos == epos) return TRUE;
     if (output_select_format(&(chk->format))) {

	  if (!output_suggest_format(&(chk->format),&fmt))
	       memcpy(&fmt,&player_fallback_format,sizeof(Dataformat));
	       
	  if (!dataformat_samples_equal(&(chk->format),&fmt)) {
	       chk = chunk_convert_sampletype(chk,&fmt);
	       b = player_play(chk,spos,epos,lp);
	       gtk_object_sink(GTK_OBJECT(chk));
	       return b;
	  }

	  user_error("The sound format of this file is not supported for "
		     "playing.");
	  return TRUE;
     }
     ch = chunk_open(chk);
     gtk_object_ref(GTK_OBJECT(chk));
     rateest_init(chk->format.samplerate);
     loopstart = spos;
     loopend = epos;
     loop = lp;
     player_bufsize = 0;
     player_bufpos = 0;
     check_small_loop();
     curpos = spos;
     realpos_offset = spos;
     player_work();
     return FALSE;
}

off_t player_get_buffer_pos(void)
{
     return curpos;
}

/* This function calculates at what sample the listener is.
 * Since for most drivers the amount of buffering is unknown, I use clock time
 * to find out where we should be. After some time has passed, we should be
 * able to estimate the buffer amount by comparing the "clock" position and
 * the buffer position. Also, we check for overruns by checking for 
 * unreasonable values */
off_t player_get_real_pos(void)
{
     off_t o,x;
     o = rateest_frames_played() + realpos_offset;
     if (loop && curpos < loopend && o >= loopend) {
	  x = o-loopstart;
	  o = loopstart + x%(loopend-loopstart);
     }      
     return o;
}

void player_get_range(off_t *startpos, off_t *endpos)
{
     *startpos = loopstart;
     *endpos = loopend;
}

void player_set_buffer_pos(off_t pos)
{
     if (pos != curpos) {
	  curpos = pos;
	  output_clear_buffers();
	  rateest_init(rateest_real_samplerate());
	  realpos_offset = pos;
	  player_bufsize = player_bufpos = 0;
	  player_work();
     }
}

void player_replace(Chunk *chunk, off_t start, off_t end, off_t pos)
{
     ChunkHandle *c;
     off_t rp = player_get_real_pos();
     off_t oldpos = curpos;

     /* Special case - we're playing a loop and the sound has looped, but 
      * because of the output delay the cursor and the audible sound hasn't 
      * come there yet. In that case, we're forced to flush buffers etc 
      * to avoid sending out the old loop to the listener. */
     if (loop && rp > oldpos) {
	  player_play(chunk,rp,end,TRUE);
	  loopstart = start;
	  return;
     }

     c = chunk_open(chunk);
     if (c == NULL) return;
     chunk_close(ch);
     gtk_object_unref(GTK_OBJECT(ch));
     ch = c;
     gtk_object_ref(GTK_OBJECT(ch));

     loopstart = start;
     loopend = end;
     curpos = pos;
     
     check_small_loop();

     /* Update timerel */
     realpos_offset += pos - oldpos;
}


gboolean player_playing(void)
{
     return (ch != NULL);
}



void player_stop(void)
{
     if (ch == NULL) return;
     chunk_close(ch);
     gtk_object_unref(GTK_OBJECT(ch));
     ch = NULL;
     output_clear_buffers();
     output_stop();
}

gboolean player_looping(void)
{
     return loop;
}

void player_nudge(gfloat seconds)
{
     off_t samps,newpos;
     samps = seconds * ch->format.samplerate;
     newpos = player_get_real_pos() + samps;
     if (newpos < 0) newpos = 0;
     player_set_buffer_pos(newpos);
}
