/*
 * Copyright (C) 2004 2005, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */

/* This module was re-written completely from scratch in Sep 2004. 
 * Thanks to Erica Andrews for contributing the original esound
 * driver. */

#include <esd.h>
#include <errno.h>
#include <unistd.h>

/* Playback/Recording connections (<= 0 if not connected) */
static int esound_play_fd = -1;
static int esound_record_fd = -1;
static Dataformat esound_play_format;

static void esound_init(void)
{
     /* Just try once here so the user gets a clear error message if ESD
      * isn't available */
     int fd;
     gchar *c;
     fd = esd_open_sound(NULL);
     if (fd < 0) {
	  c = g_strdup_printf("Couldn't connect to ESD daemon: %s",
			      strerror(errno));
	  user_error(c);
	  g_free(c);
     }
     esd_close(fd);
}

static void esound_quit(void)
{
}

static void esound_select_format(Dataformat *format, int *fdp, 
				 gboolean recording)
{
     int fmt = ESD_STREAM;
     g_assert(*fdp < 0);     
     if (recording) fmt |= ESD_RECORD; else fmt |= ESD_PLAY;
     if (format->samplesize == 1) fmt |= ESD_BITS8;
     else if (format->samplesize == 2) fmt |= ESD_BITS16;
     else return;
     if (format->channels == 1) fmt |= ESD_MONO;
     else if (format->channels == 2) fmt |= ESD_STEREO;
     else return;
     if (recording)
	  *fdp = esd_record_stream(fmt,format->samplerate,
				   NULL,"mhWaveEdit");
     else
	  *fdp = esd_play_stream(fmt,format->samplerate,NULL,
				 "mhWaveEdit");     
}

static gboolean esound_output_select_format(Dataformat *format)
{
     if (format->type == DATAFORMAT_FLOAT) return TRUE;
     esound_select_format(format,&esound_play_fd,FALSE);
     if (esound_play_fd < 0) { 
	  fputs("mhWaveEdit: Unable to open playback stream\n",stderr);
	  return TRUE;
     }
     memcpy(&esound_play_format,format,sizeof(Dataformat));
     return FALSE;
}

static gboolean esound_input_select_format(Dataformat *format)
{
     if (format->type == DATAFORMAT_FLOAT) return TRUE;
     esound_select_format(format,&esound_record_fd,TRUE);
     if (esound_record_fd < 0) {
	  fputs("mhWaveEdit: Unable to open recording stream\n",stderr);
	  return TRUE;
     }
     return FALSE;
}

static gboolean esound_output_want_data(void)
{
     fd_set wfds;
     int i;
     struct timeval t = { 0.0, 0.0 };    
     FD_ZERO(&wfds);
     FD_SET(esound_play_fd,&wfds);
     i = select(esound_play_fd+1, NULL, &wfds, NULL, &t);
     if (i == -1) {
	  perror("mhWaveEdit: EsounD driver: select");
	  return TRUE;
     }
     return (i>0);
}

static guint esound_output_play(gchar *buf, guint len)
{
     ssize_t sst;
     if (buf == NULL) return 0;
     sst = write(esound_play_fd, buf, len);
     if (sst < 0) {
	  perror("mhWaveEdit: EsounD driver: write");
	  return 0;
     }
     /* FIXME: It's theoretically possible that just part of a sample was 
      * written if the socket buffer size isn't a multiple of 4. */
     return (guint) sst;
     
}

static void esound_output_stop(void)
{
     if (esound_play_fd >= 0) esd_close(esound_play_fd);
     esound_play_fd = -1;
}

static void esound_input_stop(void)
{
     if (esound_record_fd >= 0) esd_close(esound_record_fd);
     esound_record_fd = -1;
}

static gfloat esound_output_play_min_delay_time(void)
{
     /* Could be improved.. */
     return 0.0;     
}

static off_t esound_output_played_bytes(void)
{
     /* Unused ATM */
     g_assert_not_reached();
     return 0;
}

static void esound_output_clear_buffers(void)
{
     return;
}

static gboolean esound_input_supported(void)
{
     return TRUE;
}

static void esound_input_store(Ringbuf *buffer)
{
     gchar buf[4096];
     int i;
     fd_set set;
     struct timeval tv = { 0, 0 };
     int sz;
     if (esound_record_fd < 0) return;
     while (1) {
	  sz = ringbuf_freespace(buffer);
	  if (sz > sizeof(buf)) sz = sizeof(buf);
	  if (sz == 0) return;
	  FD_ZERO(&set);
	  FD_SET(esound_record_fd, &set);
	  i = select(esound_record_fd+1, &set, NULL, NULL, &tv);
	  if (i<0)
	       perror("mhWaveEdit: EsounD driver: select");	  
	  if (i == 0) return;
	  i = read(esound_record_fd,buf,sz);
	  if (i == 0) {
	       user_error("Esound connection closed by server");
	       esound_record_fd = -1;
	       return;
	  }
	  if (i<0) {
	       if (errno == EAGAIN || errno == EINTR) continue;
	       perror("mhWaveEdit: EsounD driver: read");
	       return;
	  }
	  ringbuf_enqueue(buffer,buf,i);
     }     
}
